
# distutils: libraries = STD_LIBS
"""
A refine-by-two AMR-specific quadtree



"""



import numpy as np

cimport cython
cimport numpy as np
from cython.operator cimport dereference as deref, preincrement as inc
from libc.stdlib cimport abs, free, malloc

from yt.utilities.lib.fp_utils cimport fmax, fmin

from yt.utilities.exceptions import YTIntDomainOverflow


cdef extern from "platform_dep.h":
    # NOTE that size_t might not be int
    void *alloca(int)

cdef struct QuadTreeNode:
    np.float64_t *val
    np.float64_t weight_val
    np.int64_t pos[2]
    QuadTreeNode *children[2][2]

ctypedef void QTN_combine(QuadTreeNode *self,
        np.float64_t *val, np.float64_t weight_val,
        int nvals)

cdef void QTN_add_value(QuadTreeNode *self,
        np.float64_t *val, np.float64_t weight_val,
        int nvals):
    cdef int i
    for i in range(nvals):
        self.val[i] += val[i]
    self.weight_val += weight_val

cdef void QTN_max_value(QuadTreeNode *self,
        np.float64_t *val, np.float64_t weight_val,
        int nvals):
    cdef int i
    for i in range(nvals):
        self.val[i] = fmax(val[i], self.val[i])
    self.weight_val = 1.0

cdef void QTN_min_value(QuadTreeNode *self,
        np.float64_t *val, np.float64_t weight_val,
        int nvals):
    cdef int i
    #cdef np.float64_t *big_num = 1.0
    #big_num = 1.0 #1e10
    for i in range(nvals):
        if self.val[i] == 0:
          self.val[i] = 1e50
        # end if
        self.val[i] = fmin(val[i], self.val[i])
    self.weight_val = 1.0

cdef void QTN_refine(QuadTreeNode *self, int nvals):
    cdef int i, j
    cdef np.int64_t npos[2]
    cdef np.float64_t *tvals = <np.float64_t *> alloca(
            sizeof(np.float64_t) * nvals)
    for i in range(nvals): tvals[i] = 0.0
    for i in range(2):
        npos[0] = self.pos[0] * 2 + i
        for j in range(2):
            npos[1] = self.pos[1] * 2 + j
            # We have to be careful with allocation...
            self.children[i][j] = QTN_initialize(
                        npos, nvals, tvals, 0.0)

cdef QuadTreeNode *QTN_initialize(np.int64_t pos[2], int nvals,
                        np.float64_t *val, np.float64_t weight_val):
    cdef QuadTreeNode *node
    cdef int i, j
    node = <QuadTreeNode *> malloc(sizeof(QuadTreeNode))
    node.pos[0] = pos[0]
    node.pos[1] = pos[1]
    node.val = <np.float64_t *> malloc(
                nvals * sizeof(np.float64_t))
    for i in range(2):
        for j in range(2):
            node.children[i][j] = NULL
    if val != NULL:
        for i in range(nvals):
            node.val[i] = val[i]
        node.weight_val = weight_val
    return node

cdef void QTN_free(QuadTreeNode *node):
    cdef int i, j
    for i in range(2):
        for j in range(2):
            if node.children[i][j] == NULL: continue
            QTN_free(node.children[i][j])
    free(node.val)
    free(node)

cdef class QuadTree:
    cdef int nvals
    cdef QuadTreeNode ***root_nodes
    cdef public np.int64_t top_grid_dims[2]
    cdef int merged
    cdef int num_cells
    cdef QTN_combine *combine
    cdef np.float64_t bounds[4]
    cdef np.float64_t dds[2]
    cdef np.int64_t last_dims[2]
    cdef int max_level

    def __cinit__(self, np.ndarray[np.int64_t, ndim=1] top_grid_dims,
                  int nvals, bounds, method = "integrate"):
        if method == "integrate":
            self.combine = QTN_add_value
        elif method == "mip" or \
             method == "max":
            self.combine = QTN_max_value
        elif method == "min":
            self.combine = QTN_min_value
        else:
            raise NotImplementedError(f"Unknown projection method {self.method!r}")
        self.merged = 1
        self.max_level = 0
        cdef int i, j
        cdef np.int64_t pos[2]
        cdef np.float64_t *vals = <np.float64_t *> malloc(
                sizeof(np.float64_t)*nvals)
        cdef np.float64_t weight_val = 0.0
        self.nvals = nvals
        for i in range(nvals): vals[i] = 0.0
        for i in range(4):
            self.bounds[i] = bounds[i]

        self.top_grid_dims[0] = top_grid_dims[0]
        self.top_grid_dims[1] = top_grid_dims[1]
        self.dds[0] = (self.bounds[1] - self.bounds[0])/self.top_grid_dims[0]
        self.dds[1] = (self.bounds[3] - self.bounds[2])/self.top_grid_dims[1]

        self.root_nodes = <QuadTreeNode ***> \
            malloc(sizeof(QuadTreeNode **) * top_grid_dims[0])

        # We initialize our root values to 0.0.
        for i in range(top_grid_dims[0]):
            pos[0] = i
            self.root_nodes[i] = <QuadTreeNode **> \
                malloc(sizeof(QuadTreeNode *) * top_grid_dims[1])
            for j in range(top_grid_dims[1]):
                pos[1] = j
                self.root_nodes[i][j] = QTN_initialize(
                    pos, nvals, vals, weight_val)
        self.num_cells = self.top_grid_dims[0] * self.top_grid_dims[1]
        free(vals)

    cdef int count_total_cells(self, QuadTreeNode *root):
        cdef int total = 0
        cdef int i, j
        if root.children[0][0] == NULL: return 1
        for i in range(2):
            for j in range(2):
                total += self.count_total_cells(root.children[i][j])
        return total + 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int fill_buffer(self, QuadTreeNode *root, int curpos,
                          np.ndarray[np.int32_t, ndim=1] refined,
                          np.ndarray[np.float64_t, ndim=2] values,
                          np.ndarray[np.float64_t, ndim=1] wval):
        cdef int i, j
        for i in range(self.nvals):
            values[curpos, i] = root.val[i]
        wval[curpos] = root.weight_val
        if root.children[0][0] != NULL: refined[curpos] = 1
        else: return curpos+1
        curpos += 1
        for i in range(2):
            for j in range(2):
                curpos = self.fill_buffer(root.children[i][j], curpos,
                                 refined, values, wval)
        return curpos

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int unfill_buffer(self, QuadTreeNode *root, int curpos,
                          np.ndarray[np.int32_t, ndim=1] refined,
                          np.ndarray[np.float64_t, ndim=2] values,
                          np.ndarray[np.float64_t, ndim=1] wval):
        cdef int i, j
        for i in range(self.nvals):
            root.val[i] = values[curpos, i]
        root.weight_val = wval[curpos]
        if refined[curpos] == 0: return curpos+1
        curpos += 1
        cdef QuadTreeNode *child
        cdef np.int64_t pos[2]
        for i in range(2):
            for j in range(2):
                pos[0] = root.pos[0]*2 + i
                pos[1] = root.pos[1]*2 + j
                child = QTN_initialize(pos, self.nvals, NULL, 0.0)
                root.children[i][j] = child
                curpos = self.unfill_buffer(child, curpos, refined, values, wval)
        return curpos


    @cython.boundscheck(False)
    @cython.wraparound(False)
    def frombuffer(self, np.ndarray[np.int32_t, ndim=1] refined,
                         np.ndarray[np.float64_t, ndim=2] values,
                         np.ndarray[np.float64_t, ndim=1] wval,
                         method):
        if method == "mip" or method == "max" or method == -1:
            self.merged = -1
        if method == "min" or method == -2:
            self.merged = -2
        elif method == "integrate" or method == 1:
            self.merged = 1
        cdef int curpos = 0
        self.num_cells = wval.shape[0]
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                curpos = self.unfill_buffer(self.root_nodes[i][j], curpos,
                                 refined, values, wval)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def tobuffer(self):
        cdef int total = self.num_cells
        # We now have four buffers:
        # Refined or not (total,) int32
        # Values in each node (total, nvals) float64
        # Weight values in each node (total,) float64
        cdef np.ndarray[np.int32_t, ndim=1] refined
        refined = np.zeros(total, dtype='int32')
        cdef np.ndarray[np.float64_t, ndim=2] values
        values = np.zeros((total, self.nvals), dtype='float64')
        cdef np.ndarray[np.float64_t, ndim=1] wval
        wval = np.zeros(total, dtype='float64')
        cdef int curpos = 0
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                curpos = self.fill_buffer(self.root_nodes[i][j], curpos,
                                 refined, values, wval)
        return (refined, values, wval)

    def get_args(self):
        return (self.top_grid_dims[0], self.top_grid_dims[1], self.nvals)

    cdef int add_to_position(self,
                 int level, np.int64_t pos[2],
                 np.float64_t *val,
                 np.float64_t weight_val, int skip = 0):
        cdef int i, j, L
        cdef QuadTreeNode *node
        node = self.find_on_root_level(pos, level)
        if node == NULL:
            return -1
        if level > self.max_level:
            self.max_level = level
        for L in range(level):
            if node.children[0][0] == NULL:
                QTN_refine(node, self.nvals)
                self.num_cells += 4
            # Maybe we should use bitwise operators?
            i = (pos[0] >> (level - L - 1)) & 1
            j = (pos[1] >> (level - L - 1)) & 1
            node = node.children[i][j]
        if skip == 1: return 0
        self.combine(node, val, weight_val, self.nvals)
        return 0

    @cython.cdivision(True)
    cdef QuadTreeNode *find_on_root_level(self, np.int64_t pos[2], int level):
        # We need this because the root level won't just have four children
        # So we find on the root level, then we traverse the tree.
        cdef np.int64_t i, j
        i = pos[0] >> level
        j = pos[1] >> level
        if i >= self.top_grid_dims[0] or i < 0 or \
           j >= self.top_grid_dims[1] or j < 0:
            self.last_dims[0] = i
            self.last_dims[1] = j
            return NULL
        return self.root_nodes[i][j]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_array_to_tree(self, int level, np.ndarray[np.int64_t, ndim=1] pxs,
            np.ndarray[np.int64_t, ndim=1] pys,
            np.ndarray[np.float64_t, ndim=2] pvals,
            np.ndarray[np.float64_t, ndim=1] pweight_vals,
            int skip = 0):
        cdef int p
        cdef np.float64_t *vals
        cdef np.float64_t *data = <np.float64_t *> pvals.data
        cdef np.int64_t pos[2]
        for p in range(pxs.shape[0]):
            vals = data + self.nvals*p
            pos[0] = pxs[p]
            pos[1] = pys[p]
            self.add_to_position(level, pos, vals, pweight_vals[p], skip)
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_chunk_to_tree(self,
            np.ndarray[np.int64_t, ndim=1] pxs,
            np.ndarray[np.int64_t, ndim=1] pys,
            np.ndarray[np.int64_t, ndim=1] level,
            np.ndarray[np.float64_t, ndim=2] pvals,
            np.ndarray[np.float64_t, ndim=1] pweight_vals):
        cdef int ps = pxs.shape[0]
        cdef int p, rv
        cdef np.float64_t *vals
        cdef np.float64_t *data = <np.float64_t *> pvals.data
        cdef np.int64_t pos[2]
        for p in range(ps):
            vals = data + self.nvals*p
            pos[0] = pxs[p]
            pos[1] = pys[p]
            rv = self.add_to_position(level[p], pos, vals, pweight_vals[p])
            if rv == -1:
                raise YTIntDomainOverflow(
                    (self.last_dims[0], self.last_dims[1]),
                    (self.top_grid_dims[0], self.top_grid_dims[1]))
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def initialize_chunk(self,
            np.ndarray[np.int64_t, ndim=1] pxs,
            np.ndarray[np.int64_t, ndim=1] pys,
            np.ndarray[np.int64_t, ndim=1] level):
        cdef int num = pxs.shape[0]
        cdef int p, rv
        cdef np.int64_t pos[2]
        for p in range(num):
            pos[0] = pxs[p]
            pos[1] = pys[p]
            rv = self.add_to_position(level[p], pos, NULL, 0.0, 1)
            if rv == -1:
                raise YTIntDomainOverflow(
                    (self.last_dims[0], self.last_dims[1]),
                    (self.top_grid_dims[0], self.top_grid_dims[1]))
        return

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def get_all(self, int count_only = 0, int method = 1):
        cdef int i, j, vi
        cdef int total = 0
        self.merged = method
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                total += self.count(self.root_nodes[i][j])
        if count_only: return total
        # Allocate our array
        cdef np.ndarray[np.int64_t, ndim=1] oix
        cdef np.ndarray[np.int64_t, ndim=1] oiy
        cdef np.ndarray[np.int64_t, ndim=1] oires
        cdef np.ndarray[np.float64_t, ndim=2] nvals
        cdef np.ndarray[np.float64_t, ndim=1] nwvals
        oix = np.zeros(total, dtype='int64')
        oiy = np.zeros(total, dtype='int64')
        oires = np.zeros(total, dtype='int64')
        nvals = np.zeros( (total, self.nvals), dtype='float64')
        nwvals = np.zeros( total, dtype='float64')
        cdef np.int64_t curpos = 0
        cdef np.int64_t *ix = <np.int64_t *> oix.data
        cdef np.int64_t *iy = <np.int64_t *> oiy.data
        cdef np.int64_t *ires = <np.int64_t *> oires.data
        cdef np.float64_t *vdata = <np.float64_t *> nvals.data
        cdef np.float64_t *wdata = <np.float64_t *> nwvals.data
        cdef np.float64_t wtoadd
        cdef np.float64_t *vtoadd = <np.float64_t *> malloc(
                sizeof(np.float64_t)*self.nvals)
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                for vi in range(self.nvals): vtoadd[vi] = 0.0
                wtoadd = 0.0
                curpos += self.fill(self.root_nodes[i][j],
                    curpos, ix, iy, ires, vdata, wdata, vtoadd, wtoadd, 0)
        free(vtoadd)
        return oix, oiy, oires, nvals, nwvals

    cdef int count(self, QuadTreeNode *node):
        cdef int i, j
        cdef int count = 0
        if node.children[0][0] == NULL: return 1
        for i in range(2):
            for j in range(2):
                count += self.count(node.children[i][j])
        return count

    @cython.cdivision(True)
    cdef int fill(self, QuadTreeNode *node,
                        np.int64_t curpos,
                        np.int64_t *ix,
                        np.int64_t *iy,
                        np.int64_t *ires,
                        np.float64_t *vdata,
                        np.float64_t *wdata,
                        np.float64_t *vtoadd,
                        np.float64_t wtoadd,
                        np.int64_t level):
        cdef int i, j, n
        cdef np.float64_t *vorig
        vorig = <np.float64_t *> malloc(sizeof(np.float64_t) * self.nvals)
        if node.children[0][0] == NULL:
            if self.merged == -1:
                for i in range(self.nvals):
                    vdata[self.nvals * curpos + i] = fmax(node.val[i], vtoadd[i])
            elif self.merged == -2:
                for i in range(self.nvals):
                    vdata[self.nvals * curpos + i] = fmin(node.val[i], vtoadd[i])
                wdata[curpos] = 1.0
            else:
                for i in range(self.nvals):
                    vdata[self.nvals * curpos + i] = node.val[i] + vtoadd[i]
                wdata[curpos] = node.weight_val + wtoadd
            ires[curpos] = level
            ix[curpos] = node.pos[0]
            iy[curpos] = node.pos[1]
            return 1
        cdef np.int64_t added = 0
        if self.merged == 1:
            for i in range(self.nvals):
                vorig[i] = vtoadd[i]
                vtoadd[i] += node.val[i]
            wtoadd += node.weight_val
        elif self.merged == -1 or self.merged == -2:
            for i in range(self.nvals):
                vtoadd[i] = node.val[i]
        for i in range(2):
            for j in range(2):
                if self.merged == -1:
                    for n in range(self.nvals):
                        vtoadd[n] = node.val[n]
                added += self.fill(node.children[i][j],
                        curpos + added, ix, iy, ires, vdata, wdata,
                        vtoadd, wtoadd, level + 1)
        if self.merged == 1:
            for i in range(self.nvals):
                vtoadd[i] = vorig[i]
            wtoadd -= node.weight_val
        free(vorig)
        return added

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_image(self, np.ndarray[np.float64_t, ndim=2] buffer, _bounds,
                   int val_index = 0, int weighted = 0):
        cdef np.float64_t pos[2]
        cdef np.float64_t dds[2]
        cdef int nn[2]
        cdef int i, j
        cdef np.float64_t bounds[4]
        cdef np.float64_t opos[4]
        cdef np.float64_t weight = 0.0, value = 0.0
        cdef np.float64_t *wval = NULL
        if weighted == 1:
            wval = &weight
        for i in range(4):
            bounds[i] = _bounds[i]
        for i in range(2):
            nn[i] = buffer.shape[i]
            dds[i] = (bounds[i*2 + 1] - bounds[i*2])/nn[i]
        pos[0] = bounds[0]
        opos[0] = opos[1] = pos[0] + dds[0]
        for i in range(nn[0]):
            pos[1] = bounds[2]
            opos[2] = opos[3] = pos[1] + dds[1]
            for j in range(nn[1]):
                # We start at level zero.  In the future we could optimize by
                # retaining oct information from previous cells.
                if not ((opos[0] <= pos[0] <= opos[1]) and
                        (opos[2] <= pos[1] <= opos[3])):
                    value = self.find_value_at_pos(pos, val_index,
                                        opos, wval)
                buffer[i,j] = value
                if weighted == 1:
                    buffer[i,j] /= weight
                pos[1] += dds[1]
            pos[0] += dds[0]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef np.float64_t find_value_at_pos(self, np.float64_t pos[2],
                                         int val_index, np.float64_t opos[4],
                                         np.float64_t *wval = NULL):
        cdef int i
        cdef np.int64_t ind[2]
        cdef np.float64_t cp[2]
        cdef np.float64_t dds[2]
        cdef np.float64_t value = 0.0
        cdef np.float64_t weight = 0.0
        cdef QuadTreeNode *cur
        for i in range(2):
            ind[i] = <np.int64_t> (pos[i]/self.dds[i])
            cp[i] = (ind[i] + 0.5) * self.dds[i]
            dds[i] = self.dds[i]
        cur = self.root_nodes[ind[0]][ind[1]]
        value += cur.val[val_index]
        weight += cur.weight_val
        while cur.children[0][0] != NULL:
            for i in range(2):
                # Note that below offset by half a dx for center, after
                # updating to the next level
                dds[i] = dds[i] * 0.5
                if cp[i] >= pos[i]:
                    ind[i] = 0
                    cp[i] -= dds[i] * 0.5
                else:
                    ind[i] = 1
                    cp[i] += dds[i] * 0.5
            cur = cur.children[ind[0]][ind[1]]
            value += cur.val[val_index]
            weight += cur.weight_val
        opos[0] = cp[0] - dds[0] * 0.5
        opos[1] = cp[0] + dds[0] * 0.5
        opos[2] = cp[1] - dds[1] * 0.5
        opos[3] = cp[1] + dds[1] * 0.5
        if wval != NULL:
            wval[0] = weight
        return value

    def __dealloc__(self):
        cdef int i, j
        for i in range(self.top_grid_dims[0]):
            for j in range(self.top_grid_dims[1]):
                QTN_free(self.root_nodes[i][j])
            free(self.root_nodes[i])
        free(self.root_nodes)

cdef void QTN_merge_nodes(QuadTreeNode *n1, QuadTreeNode *n2, int nvals,
                          QTN_combine *func):
    # We have four choices when merging nodes.
    # 1. If both nodes have no refinement, then we add values of n2 to n1.
    # 2. If both have refinement, we call QTN_merge_nodes on all four children.
    # 3. If n2 has refinement and n1 does not, we detach n2's children and
    #    attach them to n1.
    # 4. If n1 has refinement and n2 does not, we add the value of n2 to n1.
    cdef int i, j

    func(n1, n2.val, n2.weight_val, nvals)
    if n1.children[0][0] == n2.children[0][0] == NULL:
        pass
    elif n1.children[0][0] != NULL and n2.children[0][0] != NULL:
        for i in range(2):
            for j in range(2):
                QTN_merge_nodes(n1.children[i][j], n2.children[i][j], nvals, func)
    elif n1.children[0][0] == NULL and n2.children[0][0] != NULL:
        for i in range(2):
            for j in range(2):
                n1.children[i][j] = n2.children[i][j]
                n2.children[i][j] = NULL
    elif n1.children[0][0] != NULL and n2.children[0][0] == NULL:
        pass
    else:
        raise RuntimeError

def merge_quadtrees(QuadTree qt1, QuadTree qt2, method = 1):
    cdef int i, j
    qt1.num_cells = 0
    cdef QTN_combine *func
    if method == 1:
        qt1.merged = 1
        func = QTN_add_value
    elif method == -1:
        qt1.merged = -1
        func = QTN_max_value
    elif method == -2:
        qt1.merged = -2
        func = QTN_min_value
    else:
        raise NotImplementedError
    if qt1.merged != 0 or qt2.merged != 0:
        assert(qt1.merged == qt2.merged)
    for i in range(qt1.top_grid_dims[0]):
        for j in range(qt1.top_grid_dims[1]):
            QTN_merge_nodes(qt1.root_nodes[i][j],
                            qt2.root_nodes[i][j],
                            qt1.nvals, func)
            qt1.num_cells += qt1.count_total_cells(
                                qt1.root_nodes[i][j])



raise RuntimeError(
    "The level_sets module has been moved to yt.data_objects.level_sets."
    "Please, change the import in your scripts from "
    "'from yt.analysis_modules.level_sets' to "
    "'from yt.data_objects.level_sets.'."
)



import re
import tempfile
from collections import namedtuple
from itertools import chain
from pathlib import Path

import pytest

from yt.frontends.ramses.data_structures import RAMSESFileSanitizer

PathTuple = namedtuple(
    "PathTuple", ("output_dir", "group_dir_name", "info_file", "paths_to_try")
)


def generate_paths(create):
    with tempfile.TemporaryDirectory() as tmpdir:
        output_dir = Path(tmpdir) / "output_00123"
        output_dir.mkdir()
        amr_file = output_dir / "amr_00123.out00001"
        if create:
            amr_file.touch()
        info_file = output_dir / "info_00123.txt"
        if create:
            info_file.touch()

        # Test with regular structure
        output_dir2 = Path(tmpdir) / "output_00124"
        output_dir2.mkdir()

        group_dir2 = output_dir2 / "group_00001"
        group_dir2.mkdir()
        info_file2 = group_dir2 / "info_00124.txt"
        if create:
            info_file2.touch()
        amr_file2 = group_dir2 / "amr_00124.out00001"
        if create:
            amr_file2.touch()

        yield (
            PathTuple(
                output_dir=output_dir,
                group_dir_name=None,
                info_file=info_file,
                paths_to_try=(output_dir, info_file, amr_file),
            ),
            PathTuple(
                output_dir=output_dir2,
                group_dir_name=group_dir2.name,
                info_file=info_file2,
                paths_to_try=(output_dir2, info_file2, group_dir2, amr_file2),
            ),
        )


@pytest.fixture
def valid_path_tuples():
    yield from generate_paths(create=True)


@pytest.fixture
def invalid_path_tuples():
    yield from generate_paths(create=False)


def test_valid_sanitizing(valid_path_tuples):
    for answer in valid_path_tuples:
        for path in answer.paths_to_try:
            sanitizer = RAMSESFileSanitizer(path)
            sanitizer.validate()

            assert sanitizer.root_folder == answer.output_dir
            assert sanitizer.group_name == answer.group_dir_name
            assert sanitizer.info_fname == answer.info_file


def test_invalid_sanitizing(valid_path_tuples, invalid_path_tuples):
    for path in chain(*(pt.paths_to_try for pt in invalid_path_tuples)):
        sanitizer = RAMSESFileSanitizer(path)

        if path.exists():
            expected_error = ValueError
            expected_error_message = (
                "Could not determine output directory from '.*'\n"
                "Expected a directory name of form .* "
                "containing an info_\\*.txt file and amr_\\* files."
            )
        else:
            expected_error = FileNotFoundError
            expected_error_message = re.escape(
                f"No such file or directory '{str(path)}'"
            )

        with pytest.raises(expected_error, match=expected_error_message):
            sanitizer.validate()

    for path in chain(*(pt.paths_to_try for pt in valid_path_tuples)):
        expected_error_message = re.escape(
            f"No such file or directory '{str(path/'does_not_exist.txt')}'"
        )
        sanitizer = RAMSESFileSanitizer(path / "does_not_exist.txt")
        with pytest.raises(FileNotFoundError, match=expected_error_message):
            sanitizer.validate()

    expected_error_message = "No such file or directory '.*'"
    sanitizer = RAMSESFileSanitizer(Path("this") / "does" / "not" / "exist")
    with pytest.raises(FileNotFoundError, match=expected_error_message):
        sanitizer.validate()



import os
import weakref
from typing import Type

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset, ParticleFile, validate_index_order
from yt.funcs import mylog, setdefaultattr
from yt.geometry.geometry_handler import Index
from yt.geometry.grid_geometry_handler import GridIndex
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.utilities.file_handler import HDF5FileHandler, warn_h5py
from yt.utilities.physical_ratios import cm_per_mpc

from .fields import FLASHFieldInfo


class FLASHGrid(AMRGridPatch):
    _id_offset = 1
    # __slots__ = ["_level_id", "stop_index"]

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level


class FLASHHierarchy(GridIndex):

    grid = FLASHGrid
    _preload_implemented = True

    def __init__(self, ds, dataset_type="flash_hdf5"):
        self.dataset_type = dataset_type
        self.field_indexes = {}
        self.dataset = weakref.proxy(ds)
        # for now, the index file is the dataset!
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self._handle = ds._handle
        self._particle_handle = ds._particle_handle
        self.float_type = np.float64
        GridIndex.__init__(self, ds, dataset_type)

    def _initialize_data_storage(self):
        pass

    def _detect_output_fields(self):
        self.field_list = [
            ("flash", s.decode("ascii", "ignore"))
            for s in self._handle["/unknown names"][:].flat
        ]
        if "/particle names" in self._particle_handle:
            self.field_list += [
                ("io", "particle_" + s[0].decode("ascii", "ignore").strip())
                for s in self._particle_handle["/particle names"][:]
            ]

    def _count_grids(self):
        try:
            self.num_grids = self.dataset._find_parameter(
                "integer", "globalnumblocks", True
            )
        except KeyError:
            try:
                self.num_grids = self._handle["simulation parameters"]["total blocks"][
                    0
                ]
            except KeyError:
                self.num_grids = self._handle["/simulation parameters"][0][0]

    def _parse_index(self):
        f = self._handle  # shortcut
        ds = self.dataset  # shortcut
        f_part = self._particle_handle  # shortcut

        # Initialize to the domain left / domain right
        ND = self.dataset.dimensionality
        DLE = self.dataset.domain_left_edge
        DRE = self.dataset.domain_right_edge
        for i in range(3):
            self.grid_left_edge[:, i] = DLE[i]
            self.grid_right_edge[:, i] = DRE[i]
        # We only go up to ND for 2D datasets
        self.grid_left_edge[:, :ND] = f["/bounding box"][:, :ND, 0]
        self.grid_right_edge[:, :ND] = f["/bounding box"][:, :ND, 1]
        # Move this to the parameter file
        try:
            nxb = ds.parameters["nxb"]
            nyb = ds.parameters["nyb"]
            nzb = ds.parameters["nzb"]
        except KeyError:
            nxb, nyb, nzb = (
                int(f["/simulation parameters"][f"n{ax}b"]) for ax in "xyz"
            )
        self.grid_dimensions[:] *= (nxb, nyb, nzb)
        try:
            self.grid_particle_count[:] = f_part["/localnp"][:][:, None]
            self._blockless_particle_count = (
                f_part["/tracer particles"].shape[0] - self.grid_particle_count.sum()
            )
        except KeyError:
            self.grid_particle_count[:] = 0.0
        self._particle_indices = np.zeros(self.num_grids + 1, dtype="int64")
        if self.num_grids > 1:
            np.add.accumulate(
                self.grid_particle_count.squeeze(), out=self._particle_indices[1:]
            )
        else:
            self._particle_indices[1] = self.grid_particle_count.squeeze()
        # This will become redundant, as _prepare_grid will reset it to its
        # current value.  Note that FLASH uses 1-based indexing for refinement
        # levels, but we do not, so we reduce the level by 1.
        self.grid_levels.flat[:] = f["/refine level"][:][:] - 1
        self.grids = np.empty(self.num_grids, dtype="object")
        for i in range(self.num_grids):
            self.grids[i] = self.grid(i + 1, self, self.grid_levels[i, 0])

        # This is a possibly slow and verbose fix, and should be re-examined!
        rdx = self.dataset.domain_width / self.dataset.domain_dimensions
        nlevels = self.grid_levels.max()
        dxs = np.ones((nlevels + 1, 3), dtype="float64")
        for i in range(nlevels + 1):
            dxs[i, :ND] = rdx[:ND] / self.dataset.refine_by**i

        if ND < 3:
            dxs[:, ND:] = rdx[ND:]

        # Because we don't care about units, we're going to operate on views.
        gle = self.grid_left_edge.ndarray_view()
        gre = self.grid_right_edge.ndarray_view()
        geom = self.dataset.geometry
        if geom != "cartesian" and ND < 3:
            if geom == "spherical" and ND < 2:
                gle[:, 1] = 0.0
                gre[:, 1] = np.pi
            gle[:, 2] = 0.0
            gre[:, 2] = 2.0 * np.pi
            return

    def _populate_grid_objects(self):
        ii = np.argsort(self.grid_levels.flat)
        gid = self._handle["/gid"][:]
        first_ind = -(self.dataset.refine_by**self.dataset.dimensionality)
        for g in self.grids[ii].flat:
            gi = g.id - g._id_offset
            # FLASH uses 1-indexed group info
            g.Children = [self.grids[i - 1] for i in gid[gi, first_ind:] if i > -1]
            for g1 in g.Children:
                g1.Parent = g
            g._prepare_grid()
            g._setup_dx()
        if self.dataset.dimensionality < 3:
            DD = self.dataset.domain_right_edge[2] - self.dataset.domain_left_edge[2]
            for g in self.grids:
                g.dds[2] = DD
        if self.dataset.dimensionality < 2:
            DD = self.dataset.domain_right_edge[1] - self.dataset.domain_left_edge[1]
            for g in self.grids:
                g.dds[1] = DD
        self.max_level = self.grid_levels.max()


class FLASHDataset(Dataset):
    _index_class: Type[Index] = FLASHHierarchy
    _field_info_class = FLASHFieldInfo
    _handle = None

    def __init__(
        self,
        filename,
        dataset_type="flash_hdf5",
        storage_filename=None,
        particle_filename=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):

        self.fluid_types += ("flash",)
        if self._handle is not None:
            return
        self._handle = HDF5FileHandler(filename)

        self.particle_filename = particle_filename

        if self.particle_filename is None:
            # try to guess the particle filename
            try:
                self._particle_handle = HDF5FileHandler(
                    filename.replace("plt_cnt", "part")
                )
                self.particle_filename = filename.replace("plt_cnt", "part")
                mylog.info(
                    "Particle file found: %s", self.particle_filename.split("/")[-1]
                )
            except OSError:
                self._particle_handle = self._handle
        else:
            # particle_filename is specified by user
            self._particle_handle = HDF5FileHandler(self.particle_filename)

        # Check if the particle file has the same time
        if self._particle_handle != self._handle:
            part_time = self._particle_handle.handle.get("real scalars")[0][1]
            plot_time = self._handle.handle.get("real scalars")[0][1]
            if not np.isclose(part_time, plot_time):
                self._particle_handle = self._handle
                mylog.warning(
                    "%s and %s are not at the same time. "
                    "This particle file will not be used.",
                    self.particle_filename,
                    filename,
                )

        # These should be explicitly obtained from the file, but for now that
        # will wait until a reorganization of the source tree and better
        # generalization.
        self.refine_by = 2

        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        self.storage_filename = storage_filename

        self.parameters["HydroMethod"] = "flash"  # always PPM DE
        self.parameters["Time"] = 1.0  # default unit is 1...

    def _set_code_unit_attributes(self):

        if "unitsystem" in self.parameters:
            # Some versions of FLASH inject quotes in the runtime parameters
            # See issue #1721
            us = self["unitsystem"].replace("'", "").replace('"', "").lower()
            if us == "cgs":
                b_factor = 1.0
            elif us == "si":
                b_factor = np.sqrt(4 * np.pi / 1e7)
            elif us == "none":
                b_factor = np.sqrt(4 * np.pi)
            else:
                raise RuntimeError(
                    "Runtime parameter unitsystem with "
                    "value %s is unrecognized" % self["unitsystem"]
                )
        else:
            b_factor = 1.0
        if self.cosmological_simulation == 1:
            length_factor = 1.0 / (1.0 + self.current_redshift)
            temperature_factor = 1.0 / (1.0 + self.current_redshift) ** 2
        else:
            length_factor = 1.0
            temperature_factor = 1.0

        setdefaultattr(self, "magnetic_unit", self.quan(b_factor, "gauss"))
        setdefaultattr(self, "length_unit", self.quan(length_factor, "cm"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "g"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        setdefaultattr(self, "velocity_unit", self.quan(1.0, "cm/s"))
        setdefaultattr(self, "temperature_unit", self.quan(temperature_factor, "K"))

    def set_code_units(self):
        super().set_code_units()

    def _find_parameter(self, ptype, pname, scalar=False):
        nn = "/{} {}".format(
            ptype, {False: "runtime parameters", True: "scalars"}[scalar]
        )
        if nn not in self._handle:
            raise KeyError(nn)
        for tpname, pval in zip(
            self._handle[nn][:, "name"], self._handle[nn][:, "value"]
        ):
            if tpname.decode("ascii", "ignore").strip() == pname:
                if hasattr(pval, "decode"):
                    pval = pval.decode("ascii", "ignore")
                if ptype == "string":
                    return pval.strip()
                else:
                    return pval
        raise KeyError(pname)

    def _parse_parameter_file(self):
        if "file format version" in self._handle:
            self._flash_version = int(self._handle["file format version"][:])
        elif "sim info" in self._handle:
            self._flash_version = int(
                self._handle["sim info"][:]["file format version"]
            )
        else:
            raise RuntimeError("Can't figure out FLASH file version.")
        # First we load all of the parameters
        hns = ["simulation parameters"]
        # note the ordering here is important: runtime parameters should
        # overwrite scalars with the same name.
        for ptype in ["scalars", "runtime parameters"]:
            for vtype in ["integer", "real", "logical", "string"]:
                hns.append(f"{vtype} {ptype}")
        if self._flash_version > 7:
            for hn in hns:
                if hn not in self._handle:
                    continue
                for varname, val in zip(
                    self._handle[hn][:, "name"], self._handle[hn][:, "value"]
                ):
                    vn = varname.strip()
                    if hn.startswith("string"):
                        pval = val.strip()
                    else:
                        pval = val
                    if vn in self.parameters and self.parameters[vn] != pval:
                        mylog.info(
                            "%s %s overwrites a simulation scalar of the same name",
                            hn[:-1],
                            vn,
                        )
                    if hasattr(pval, "decode"):
                        pval = pval.decode("ascii", "ignore")
                    self.parameters[vn.decode("ascii", "ignore")] = pval
        if self._flash_version == 7:
            for hn in hns:
                if hn not in self._handle:
                    continue
                if hn == "simulation parameters":
                    zipover = (
                        (name, self._handle[hn][name][0])
                        for name in self._handle[hn].dtype.names
                    )
                else:
                    zipover = zip(
                        self._handle[hn][:, "name"], self._handle[hn][:, "value"]
                    )
                for varname, val in zipover:
                    vn = varname.strip()
                    if hasattr(vn, "decode"):
                        vn = vn.decode("ascii", "ignore")
                    if hn.startswith("string"):
                        pval = val.strip()
                    else:
                        pval = val
                    if vn in self.parameters and self.parameters[vn] != pval:
                        mylog.info(
                            "%s %s overwrites a simulation scalar of the same name",
                            hn[:-1],
                            vn,
                        )
                    if hasattr(pval, "decode"):
                        pval = pval.decode("ascii", "ignore")
                    self.parameters[vn] = pval

        # Determine block size
        try:
            nxb = self.parameters["nxb"]
            nyb = self.parameters["nyb"]
            nzb = self.parameters["nzb"]
        except KeyError:
            nxb, nyb, nzb = (
                int(self._handle["/simulation parameters"][f"n{ax}b"]) for ax in "xyz"
            )  # FLASH2 only!

        # Determine dimensionality
        try:
            dimensionality = self.parameters["dimensionality"]
        except KeyError:
            dimensionality = 3
            if nzb == 1:
                dimensionality = 2
            if nyb == 1:
                dimensionality = 1
            if dimensionality < 3:
                mylog.warning("Guessing dimensionality as %s", dimensionality)

        self.dimensionality = dimensionality

        self.geometry = self.parameters["geometry"]
        # Determine base grid parameters
        if "lrefine_min" in self.parameters.keys():  # PARAMESH
            nblockx = self.parameters["nblockx"]
            nblocky = self.parameters["nblocky"]
            nblockz = self.parameters["nblockz"]
        else:  # Uniform Grid
            nblockx = self.parameters["iprocs"]
            nblocky = self.parameters["jprocs"]
            nblockz = self.parameters["kprocs"]

        # In case the user wasn't careful
        if self.dimensionality <= 2:
            nblockz = 1
        if self.dimensionality == 1:
            nblocky = 1

        # Determine domain boundaries
        dle = np.array([self.parameters[f"{ax}min"] for ax in "xyz"]).astype("float64")
        dre = np.array([self.parameters[f"{ax}max"] for ax in "xyz"]).astype("float64")
        if self.dimensionality < 3:
            for d in range(self.dimensionality, 3):
                if dle[d] == dre[d]:
                    mylog.warning(
                        "Identical domain left edge and right edges "
                        "along dummy dimension (%i), attempting to read anyway",
                        d,
                    )
                    dre[d] = dle[d] + 1.0
        if self.dimensionality < 3 and self.geometry == "cylindrical":
            mylog.warning("Extending theta dimension to 2PI + left edge.")
            dre[2] = dle[2] + 2 * np.pi
        elif self.dimensionality < 3 and self.geometry == "polar":
            mylog.warning("Extending theta dimension to 2PI + left edge.")
            dre[1] = dle[1] + 2 * np.pi
        elif self.dimensionality < 3 and self.geometry == "spherical":
            mylog.warning("Extending phi dimension to 2PI + left edge.")
            dre[2] = dle[2] + 2 * np.pi
        if self.dimensionality == 1 and self.geometry == "spherical":
            mylog.warning("Extending theta dimension to PI + left edge.")
            dre[1] = dle[1] + np.pi
        self.domain_left_edge = dle
        self.domain_right_edge = dre
        self.domain_dimensions = np.array([nblockx * nxb, nblocky * nyb, nblockz * nzb])

        # Try to determine Gamma
        try:
            self.gamma = self.parameters["gamma"]
        except Exception:
            mylog.info("Cannot find Gamma")
            pass

        # Get the simulation time
        self.current_time = self.parameters["time"]

        # Determine if this is a periodic box
        p = [
            self.parameters.get(f"{ax}l_boundary_type", None) == "periodic"
            for ax in "xyz"
        ]
        self._periodicity = tuple(p)

        # Determine cosmological parameters.
        try:
            self.parameters["usecosmology"]
            self.cosmological_simulation = 1
            self.current_redshift = 1.0 / self.parameters["scalefactor"] - 1.0
            self.omega_lambda = self.parameters["cosmologicalconstant"]
            self.omega_matter = self.parameters["omegamatter"]
            self.hubble_constant = self.parameters["hubbleconstant"]
            self.hubble_constant *= cm_per_mpc * 1.0e-5 * 1.0e-2  # convert to 'h'
        except Exception:
            self.current_redshift = 0.0
            self.omega_lambda = 0.0
            self.omega_matter = 0.0
            self.hubble_constant = 0.0
            self.cosmological_simulation = 0

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        try:
            fileh = HDF5FileHandler(filename)
            if "bounding box" in fileh["/"].keys():
                return True
        except (OSError, ImportError):
            pass
        return False

    @classmethod
    def _guess_candidates(cls, base, directories, files):
        candidates = [
            _ for _ in files if ("_hdf5_plt_cnt_" in _) or ("_hdf5_chk_" in _)
        ]
        # Typically, Flash won't have nested outputs.
        return candidates, (len(candidates) == 0)

    def close(self):
        self._handle.close()


class FLASHParticleFile(ParticleFile):
    pass


class FLASHParticleDataset(FLASHDataset):
    _index_class = ParticleIndex
    filter_bbox = False
    _file_class = FLASHParticleFile

    def __init__(
        self,
        filename,
        dataset_type="flash_particle_hdf5",
        storage_filename=None,
        units_override=None,
        index_order=None,
        index_filename=None,
        unit_system="cgs",
    ):
        self.index_order = validate_index_order(index_order)
        self.index_filename = index_filename

        if self._handle is not None:
            return
        self._handle = HDF5FileHandler(filename)
        self.refine_by = 2
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = storage_filename

    def _parse_parameter_file(self):
        # Let the superclass do all the work but then
        # fix the domain dimensions
        super()._parse_parameter_file()
        domain_dimensions = np.zeros(3, "int32")
        domain_dimensions[: self.dimensionality] = 1
        self.domain_dimensions = domain_dimensions
        self.filename_template = self.parameter_filename
        self.file_count = 1

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        warn_h5py(filename)
        try:
            fileh = HDF5FileHandler(filename)
            if (
                "bounding box" not in fileh["/"].keys()
                and "localnp" in fileh["/"].keys()
            ):
                return True
        except (OSError, ImportError):
            pass
        return False

    @classmethod
    def _guess_candidates(cls, base, directories, files):
        candidates = [_ for _ in files if "_hdf5_part_" in _]
        # Typically, Flash won't have nested outputs.
        return candidates, (len(candidates) == 0)



import tempfile

from yt.testing import fake_particle_ds, requires_module


@requires_module("Firefly")
def test_firefly_JSON_string():

    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        None,
        velocity_units="cm/s",
        coordinate_units="cm",
    )

    reader.dumpToJSON(write_jsons_to_disk=False)

    ## reader.JSON was not output to string correctly
    ##  either Firefly is damaged or needs a hotfix-- try reinstalling.
    ##  if that doesn't work contact the developers
    ##  at github.com/ageller/Firefly/issues.
    assert len(reader.JSON) > 0


@requires_module("Firefly")
def test_firefly_write_to_disk():
    tmpdir = tempfile.mkdtemp()

    ds = fake_particle_ds()
    ad = ds.all_data()
    reader = ad.create_firefly_object(
        tmpdir,
        velocity_units="cm/s",
        coordinate_units="cm",
    )

    reader.dumpToJSON()



"""
Bit array functions



"""


import numpy as np

cimport cython
cimport numpy as np


cdef inline void ba_set_value(np.uint8_t *buf, np.uint64_t ind,
                              np.uint8_t val) nogil:
    # This assumes 8 bit buffer
    if val > 0:
        buf[ind >> 3] |= (1 << (ind & 7))
    else:
        buf[ind >> 3] &= ~(1 << (ind & 7))

cdef inline np.uint8_t ba_get_value(np.uint8_t *buf, np.uint64_t ind) nogil:
    cdef np.uint8_t rv = (buf[ind >> 3] & (1 << (ind & 7)))
    if rv == 0: return 0
    return 1

cdef class bitarray:
    cdef np.uint8_t *buf
    cdef np.uint64_t size
    cdef np.uint64_t buf_size # Not exactly the same
    cdef public object ibuf

    cdef void _set_value(self, np.uint64_t ind, np.uint8_t val)
    cdef np.uint8_t _query_value(self, np.uint64_t ind)
    #cdef void set_range(self, np.uint64_t ind, np.uint64_t count, int val)
    #cdef int query_range(self, np.uint64_t ind, np.uint64_t count, int *val)



import os
import tempfile
from unittest import mock

import numpy as np

from yt.testing import assert_equal, assert_rel_equal, fake_amr_ds, fake_random_ds
from yt.units.unit_object import Unit

LENGTH_UNIT = 2.0


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def teardown_func(fns):
    for fn in fns:
        try:
            os.remove(fn)
        except OSError:
            pass


@mock.patch("yt.visualization._mpl_imports.FigureCanvasAgg.print_figure")
def test_projection(pf):
    fns = []
    for nprocs in [8, 1]:
        # We want to test both 1 proc and 8 procs, to make sure that
        # parallelism isn't broken
        fields = ("density", "temperature", "velocity_x", "velocity_y", "velocity_z")
        units = ("g/cm**3", "K", "cm/s", "cm/s", "cm/s")
        ds = fake_random_ds(
            64, fields=fields, units=units, nprocs=nprocs, length_unit=LENGTH_UNIT
        )
        dims = ds.domain_dimensions
        xn, yn, zn = ds.domain_dimensions
        xi, yi, zi = ds.domain_left_edge.to_ndarray() + 1.0 / (ds.domain_dimensions * 2)
        xf, yf, zf = ds.domain_right_edge.to_ndarray() - 1.0 / (
            ds.domain_dimensions * 2
        )
        dd = ds.all_data()
        coords = np.mgrid[xi : xf : xn * 1j, yi : yf : yn * 1j, zi : zf : zn * 1j]
        uc = [np.unique(c) for c in coords]
        # test if projections inherit the field parameters of their data sources
        dd.set_field_parameter("bulk_velocity", np.array([0, 1, 2]))
        proj = ds.proj(("gas", "density"), 0, data_source=dd)
        assert_equal(
            dd.field_parameters["bulk_velocity"], proj.field_parameters["bulk_velocity"]
        )

        # Some simple projection tests with single grids
        for ax, an in enumerate("xyz"):
            xax = ds.coordinates.x_axis[ax]
            yax = ds.coordinates.y_axis[ax]
            for wf in [("gas", "density"), None]:
                proj = ds.proj(
                    [("index", "ones"), ("gas", "density")], ax, weight_field=wf
                )
                if wf is None:
                    assert_equal(
                        proj[("index", "ones")].sum(),
                        LENGTH_UNIT * proj[("index", "ones")].size,
                    )
                    assert_equal(proj[("index", "ones")].min(), LENGTH_UNIT)
                    assert_equal(proj[("index", "ones")].max(), LENGTH_UNIT)
                else:
                    assert_equal(
                        proj[("index", "ones")].sum(), proj[("index", "ones")].size
                    )
                    assert_equal(proj[("index", "ones")].min(), 1.0)
                    assert_equal(proj[("index", "ones")].max(), 1.0)
                assert_equal(np.unique(proj["px"]), uc[xax])
                assert_equal(np.unique(proj["py"]), uc[yax])
                assert_equal(np.unique(proj["pdx"]), 1.0 / (dims[xax] * 2.0))
                assert_equal(np.unique(proj["pdy"]), 1.0 / (dims[yax] * 2.0))
                plots = [proj.to_pw(fields=("gas", "density")), proj.to_pw()]
                for pw in plots:
                    for p in pw.plots.values():
                        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
                        os.close(tmpfd)
                        p.save(name=tmpname)
                        fns.append(tmpname)
                frb = proj.to_frb((1.0, "unitary"), 64)
                for proj_field in [
                    ("index", "ones"),
                    ("gas", "density"),
                    ("gas", "temperature"),
                ]:
                    fi = ds._get_field_info(proj_field)
                    assert_equal(frb[proj_field].info["data_source"], proj.__str__())
                    assert_equal(frb[proj_field].info["axis"], ax)
                    assert_equal(frb[proj_field].info["field"], str(proj_field))
                    field_unit = Unit(fi.units)
                    if wf is not None:
                        assert_equal(
                            frb[proj_field].units,
                            Unit(field_unit, registry=ds.unit_registry),
                        )
                    else:
                        if frb[proj_field].units.is_code_unit:
                            proj_unit = "code_length"
                        else:
                            proj_unit = "cm"
                        if field_unit != "" and field_unit != Unit():
                            proj_unit = f"({field_unit}) * {proj_unit}"
                        assert_equal(
                            frb[proj_field].units,
                            Unit(proj_unit, registry=ds.unit_registry),
                        )
                    assert_equal(frb[proj_field].info["xlim"], frb.bounds[:2])
                    assert_equal(frb[proj_field].info["ylim"], frb.bounds[2:])
                    assert_equal(frb[proj_field].info["center"], proj.center)
                    if wf is None:
                        assert_equal(frb[proj_field].info["weight_field"], wf)
                    else:
                        assert_equal(
                            frb[proj_field].info["weight_field"],
                            proj.data_source._determine_fields(wf)[0],
                        )
            # wf == None
            assert_equal(wf, None)
            v1 = proj[("gas", "density")].sum()
            v2 = (dd[("gas", "density")] * dd[("index", f"d{an}")]).sum()
            assert_rel_equal(v1, v2.in_units(v1.units), 10)

        # Test moment projections
        def make_vsq_field(aname):
            def _vsquared(field, data):
                return data["gas", f"velocity_{aname}"] ** 2

            return _vsquared

        for ax, an in enumerate("xyz"):
            ds.add_field(
                ("gas", f"velocity_{an}_squared"),
                make_vsq_field(an),
                sampling_type="local",
                units="cm**2/s**2",
            )
            proj1 = ds.proj(
                [("gas", f"velocity_{an}"), ("gas", f"velocity_{an}_squared")],
                ax,
                weight_field=("gas", "density"),
                moment=1,
            )
            proj2 = ds.proj(
                ("gas", f"velocity_{an}"), ax, weight_field=("gas", "density"), moment=2
            )
            assert_rel_equal(
                np.sqrt(
                    proj1["gas", f"velocity_{an}_squared"]
                    - proj1["gas", f"velocity_{an}"] ** 2
                ),
                proj2["gas", f"velocity_{an}"],
                10,
            )
    teardown_func(fns)


def test_max_level():
    ds = fake_amr_ds(fields=[("gas", "density")], units=["mp/cm**3"])
    proj = ds.proj(("gas", "density"), 2, method="max", max_level=2)
    assert proj[("index", "grid_level")].max() == 2

    proj = ds.proj(("gas", "density"), 2, method="max")
    assert proj[("index", "grid_level")].max() == ds.index.max_level


def test_min_level():
    ds = fake_amr_ds(fields=[("gas", "density")], units=["mp/cm**3"])
    proj = ds.proj(("gas", "density"), 2, method="min")
    assert proj[("index", "grid_level")].min() == 0

    proj = ds.proj(("gas", "density"), 2, method="max")
    assert proj[("index", "grid_level")].min() == ds.index.min_level



from nose.plugins.attrib import attr

from yt.testing import ANSWER_TEST_TAG, fake_random_ds
from yt.utilities.answer_testing.framework import GenericImageTest
from yt.visualization.api import ProfilePlot


@attr(ANSWER_TEST_TAG)
def test_profileplot_set_axis_properties():
    ds = fake_random_ds(16)

    def create_image(filename_prefix):
        disk = ds.disk(ds.domain_center, [0.0, 0.0, 1.0], (10, "m"), (1, "m"))
        p = ProfilePlot(disk, ("gas", "density"), [("gas", "velocity_x")])
        p.save(f"{filename_prefix}_defaults")

        p.set_unit(("gas", "density"), "kg/cm**3")
        p.save(f"{filename_prefix}_xunit")

        p.set_log(("gas", "density"), False)
        p.save(f"{filename_prefix}_xunit_xlin")

        p.set_unit(("gas", "velocity_x"), "mile/hour")
        p.save(f"{filename_prefix}_xunit_xlin_yunit")

    test = GenericImageTest(ds, create_image, 12)
    test.prefix = "test_profileplot_set_axis_properties"
    test.answer_name = "profileplot_set_axis_properties"
    yield test



# distutils: include_dirs = LIB_DIR
# distutils: libraries = STD_LIBS
cimport cython
cimport libc.math as math
cimport numpy as np

import numpy as np


cdef np.float64_t h_eos4(np.float64_t kT, np.float64_t g):
    cdef np.float64_t x
    x = 2.25 * kT * kT
    return 2.5 * kT + x / (1.0 + math.sqrt(x + 1.0))

cdef np.float64_t h_eos(np.float64_t kT, np.float64_t g):
    return g * kT / (g - 1.0)

cdef np.float64_t gamma_eos(np.float64_t kT, np.float64_t g):
    return g

cdef np.float64_t gamma_eos4(np.float64_t kT, np.float64_t g):
    cdef np.float64_t x, c_p, c_v
    x = 2.25 * kT / math.sqrt(2.25 * kT * kT + 1.0)
    c_p = 2.5 + x
    c_v = 1.5 + x
    return c_p / c_v

cdef np.float64_t cs_eos4(np.float64_t kT, np.float64_t c, np.float64_t g):
    cdef np.float64_t hp, cs2;
    hp = h_eos4(kT, 0.0) + 1.0
    cs2 = kT / (3.0 * hp)
    cs2 *= (5.0 * hp - 8.0 * kT) / (hp - kT)
    return c * math.sqrt(cs2)

cdef np.float64_t cs_eos(np.float64_t kT, np.float64_t c, np.float64_t g):
    cdef np.float64_t hp, cs2;
    hp = h_eos(kT, g) + 1.0
    cs2 = g / hp * kT
    return c * math.sqrt(cs2)

ctypedef np.float64_t (*f2_type)(np.float64_t, np.float64_t)
ctypedef np.float64_t (*f3_type)(np.float64_t, np.float64_t, np.float64_t)

cdef class SRHDFields:
    cdef f2_type h
    cdef f2_type gamma
    cdef f3_type cs
    cdef np.float64_t _gamma, _c, _c2

    def __init__(self, int eos, np.float64_t gamma, np.float64_t clight):
        self._gamma = gamma
        self._c = clight
        self._c2 = clight * clight
        # Select aux functions based on eos no.
        if (eos == 4):
            self.h = h_eos4
            self.gamma = gamma_eos4
            self.cs = cs_eos4
        else:
            self.h = h_eos
            self.gamma = gamma_eos
            self.cs = cs_eos

    cdef np.float64_t _lorentz_factor(
            self,
            np.float64_t rho,
            np.float64_t mx,
            np.float64_t my,
            np.float64_t mz,
            np.float64_t kT,
        ):
        cdef np.float64_t u2, c2, vx, vy, vz
        cdef np.float64_t htilde, fac;

        fac = (1.0 / (rho * (self.h(kT, self._gamma) + 1.0))) ** 2
        u2 = (mx * mx + my * my + mz * mz) * fac
        return math.sqrt(1.0 + u2 / self._c2)

    cdef np.float64_t _four_vel(
        self,
        np.float64_t rho,
        np.float64_t kT,
        np.float64_t mi
    ):
        return mi / (rho * (self.h(kT, self._gamma) + 1.0))

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def lorentz_factor(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()

        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf
        cdef int i

        for i in range(outp.shape[0]):
            outp[i] = self._lorentz_factor(rho[i], mx[i], my[i], mz[i], kT[i])
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def gamma_field(self, temp):
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(kT)
        cdef np.float64_t[:] outp = out.ravel()
        cdef int i

        for i in range(outp.shape[0]):
            outp[i] = self.gamma(kT[i], self._gamma)
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def sound_speed(self, temp):
        cdef np.float64_t[:] kT = temp.ravel()
        out = np.empty_like(kT)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t hp, cs2;

        cdef int i
        for i in range(outp.shape[0]):
            outp[i] = self.cs(kT[i], self._c, self._gamma)
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def four_velocity_xyz(self, mom, dens, temp):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mi = mom.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t[:] ui
        cdef int i

        for i in range(outp.shape[0]):
            outp[i] = self._four_vel(rho[i], kT[i], mi[i])
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def velocity_xyz(self, dens, momx, momy, momz, temp, momi):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()
        cdef np.float64_t[:] mi = momi.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf
        cdef int i

        for i in range(outp.shape[0]):
            lf = self._lorentz_factor(rho[i], mx[i], my[i], mz[i], kT[i])
            outp[i] = self._four_vel(rho[i], kT[i], mi[i]) / lf
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def density(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()

        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf
        cdef int i

        for i in range(outp.shape[0]):
            lf = self._lorentz_factor(rho[i], mx[i], my[i], mz[i], kT[i])
            outp[i] = rho[i] / lf
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def pressure(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf
        cdef int i

        for i in range(outp.shape[0]):
            lf = self._lorentz_factor(rho[i], mx[i], my[i], mz[i], kT[i])
            outp[i] = rho[i] / lf * self._c2 * kT[i]
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def specific_thermal_energy(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf, p, ht
        cdef int i

        for i in range(outp.shape[0]):
            outp[i] = self._c2 * (self.h(kT[i], self._gamma)- kT[i])
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def kinetic_energy_density(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf, u2, hp, ux, uy, uz
        cdef int i

        for i in range(outp.shape[0]):
            hp = self.h(kT[i], self._gamma) + 1.0
            ux = self._four_vel(rho[i], kT[i], mx[i])
            uy = self._four_vel(rho[i], kT[i], my[i])
            uz = self._four_vel(rho[i], kT[i], mz[i])
            u2 = ux**2 + uy**2 + uz**2
            lf = self._lorentz_factor(rho[i], mx[i], my[i], mz[i], kT[i])
            gm1 = u2 / self._c2 / (lf + 1.0)
            p = rho[i] / lf * self._c2 * kT[i]
            outp[i] = gm1 * (rho[i] * hp * self._c2 + p)
        return out

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def mach_number(self, dens, momx, momy, momz, temp):
        cdef np.float64_t[:] rho = dens.ravel()
        cdef np.float64_t[:] mx = momx.ravel()
        cdef np.float64_t[:] my = momy.ravel()
        cdef np.float64_t[:] mz = momz.ravel()
        cdef np.float64_t[:] kT = temp.ravel()

        out = np.empty_like(dens)
        cdef np.float64_t[:] outp = out.ravel()
        cdef np.float64_t lf, u2, hp, cs
        cdef int i

        for i in range(outp.shape[0]):
            cs = self.cs(kT[i], self._c, self._gamma)
            us = cs / math.sqrt(1.0 - cs**2 / self._c2)
            u = math.sqrt(mx[i]**2 + my[i]**2 + mz[i]**2) / (rho[i] * (self.h(kT[i], self._gamma) + 1.0))
            outp[i] = u / us
        return out



from yt import load
from yt.testing import fake_amr_ds, fake_hexahedral_ds, requires_file


def do_field_type(ft):
    assert dir(ft) == sorted(dir(ft))
    assert sorted(dir(ft)) == sorted(f.name[1] for f in ft)
    for field_name in dir(ft):
        f = getattr(ft, field_name)
        assert (ft.field_type, field_name) == f.name
    for field in ft:
        f = getattr(ft, field.name[1])
        assert f == field
        assert f in ft
        assert f.name in ft
        assert f.name[1] in ft


enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"


@requires_file(enzotiny)
def test_field_name_container():
    ds = load(enzotiny)
    assert dir(ds.fields) == sorted(dir(ds.fields))
    assert sorted(ft.field_type for ft in ds.fields) == sorted(dir(ds.fields))
    for field_type in dir(ds.fields):
        assert field_type in ds.fields
        ft = getattr(ds.fields, field_type)
        do_field_type(ft)
    for field_type in ds.fields:
        assert field_type in ds.fields
        do_field_type(field_type)


def test_vertex_fields_only_in_unstructured_ds():
    def get_vertex_fields(ds):
        return [(ft, fn) for ft, fn in ds.derived_field_list if "vertex" in fn]

    ds = fake_amr_ds()
    vertex_fields = get_vertex_fields(ds)
    assert not vertex_fields

    ds = fake_hexahedral_ds()
    actual = get_vertex_fields(ds)
    expected = [
        ("all", "vertex_x"),
        ("all", "vertex_y"),
        ("all", "vertex_z"),
        ("connect1", "vertex_x"),
        ("connect1", "vertex_y"),
        ("connect1", "vertex_z"),
        ("index", "vertex_x"),
        ("index", "vertex_y"),
        ("index", "vertex_z"),
    ]
    assert actual == expected



from yt.frontends.nc4_cm1.api import CM1Dataset
from yt.testing import assert_equal, requires_file, units_override_check
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    GridValuesTest,
    can_run_ds,
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields = (("cm1", "thrhopert"), ("cm1", "zvort"))
cm1sim = "cm1_tornado_lofs/nc4_cm1_lofs_tornado_test.nc"


@requires_ds(cm1sim)
def test_cm1_mesh_fields():
    ds = data_dir_load(cm1sim)
    assert_equal(str(ds), "nc4_cm1_lofs_tornado_test.nc")

    # run the small_patch_amr tests on safe fields
    ic = ds.domain_center
    for test in small_patch_amr(ds, _fields, input_center=ic, input_weight=None):
        test_cm1_mesh_fields.__name__ = test.description
        yield test

    # manually run the Grid and Field Values tests on dbz (do not want to run the
    # ProjectionValuesTest for this field)
    if can_run_ds(ds):
        dso = [None, ("sphere", (ic, (0.1, "unitary")))]
        for field in [("cm1", "dbz")]:
            yield GridValuesTest(ds, field)
            for dobj_name in dso:
                yield FieldValuesTest(ds, field, dobj_name)


@requires_file(cm1sim)
def test_CM1Dataset():
    assert isinstance(data_dir_load(cm1sim), CM1Dataset)


@requires_file(cm1sim)
def test_units_override():
    units_override_check(cm1sim)


@requires_file(cm1sim)
def test_dims_and_meta():
    ds = data_dir_load(cm1sim)

    known_dims = ["time", "zf", "zh", "yf", "yh", "xf", "xh"]
    dims = ds.parameters["dimensions"]

    ## check the file for 2 grids and a time dimension -
    ## (time, xf, xh, yf, yh, zf, zh). The dimensions ending in
    ## f are the staggered velocity grid components and the
    ## dimensions ending in h are the scalar grid components
    assert_equal(len(dims), len(known_dims))
    for kdim in known_dims:
        assert kdim in dims

    ## check the simulation time
    assert_equal(ds.current_time, 5500.0)


@requires_file(cm1sim)
def test_if_cm1():
    ds = data_dir_load(cm1sim)
    assert float(ds.parameters["lofs_version"]) >= 1.0



from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer

m_units = "g"
p_units = "cm"
v_units = "cm / s"
r_units = "cm"

_particle_fields: KnownFieldsT = (
    ("particle_identifier", ("", [], None)),
    ("particle_position_x", (p_units, [], None)),
    ("particle_position_y", (p_units, [], None)),
    ("particle_position_z", (p_units, [], None)),
    ("particle_velocity_x", (v_units, [], None)),
    ("particle_velocity_y", (v_units, [], None)),
    ("particle_velocity_z", (v_units, [], None)),
    ("particle_mass", (m_units, [], "Virial Mass")),
    ("virial_radius", (r_units, [], "Virial Radius")),
)


class YTHaloCatalogFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    known_particle_fields = _particle_fields


class YTHaloCatalogHaloFieldInfo(FieldInfoContainer):
    known_other_fields = ()

    known_particle_fields = _particle_fields + (("ids", ("", ["member_ids"], None)),)



import os
import shutil
import tempfile
from unittest import TestCase

import numpy as np

from yt.testing import fake_random_ds
from yt.visualization.volume_rendering.api import Scene, create_volume_source


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


class LensTest(TestCase):
    # This toggles using a temporary directory. Turn off to examine images.
    use_tmpdir = True

    def setUp(self):
        if self.use_tmpdir:
            self.curdir = os.getcwd()
            # Perform I/O in safe place instead of yt main dir
            self.tmpdir = tempfile.mkdtemp()
            os.chdir(self.tmpdir)
        else:
            self.curdir, self.tmpdir = None, None

        self.field = ("gas", "density")
        self.ds = fake_random_ds(32, fields=(self.field,), units=("g/cm**3",))
        self.ds.index

    def tearDown(self):
        if self.use_tmpdir:
            os.chdir(self.curdir)
            shutil.rmtree(self.tmpdir)

    def test_perspective_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type="perspective")
        cam.position = self.ds.arr(np.array([1.0, 1.0, 1.0]), "code_length")
        vol = create_volume_source(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_perspective_{self.field[1]}.png", sigma_clip=6.0)

    def test_stereoperspective_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type="stereo-perspective")
        cam.resolution = [256, 128]
        cam.position = self.ds.arr(np.array([0.7, 0.7, 0.7]), "code_length")
        vol = create_volume_source(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_stereoperspective_{self.field[1]}.png", sigma_clip=6.0)

    def test_fisheye_lens(self):
        dd = self.ds.sphere(self.ds.domain_center, self.ds.domain_width[0] / 10)
        sc = Scene()
        cam = sc.add_camera(dd, lens_type="fisheye")
        cam.lens.fov = 360.0
        cam.set_width(self.ds.domain_width)
        v, c = self.ds.find_max(("gas", "density"))
        cam.set_position(c - 0.0005 * self.ds.domain_width)
        vol = create_volume_source(dd, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_fisheye_{self.field[1]}.png", sigma_clip=6.0)

    def test_plane_lens(self):
        dd = self.ds.sphere(self.ds.domain_center, self.ds.domain_width[0] / 10)
        sc = Scene()
        cam = sc.add_camera(dd, lens_type="plane-parallel")
        cam.set_width(self.ds.domain_width * 1e-2)
        v, c = self.ds.find_max(("gas", "density"))
        vol = create_volume_source(dd, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_plane_{self.field[1]}.png", sigma_clip=6.0)

    def test_spherical_lens(self):
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type="spherical")
        cam.resolution = [256, 128]
        cam.position = self.ds.arr(np.array([0.6, 0.5, 0.5]), "code_length")
        vol = create_volume_source(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_spherical_{self.field[1]}.png", sigma_clip=6.0)

    def test_stereospherical_lens(self):
        w = (self.ds.domain_width).in_units("code_length")
        w = self.ds.arr(w, "code_length")
        sc = Scene()
        cam = sc.add_camera(self.ds, lens_type="stereo-spherical")
        cam.resolution = [256, 256]
        cam.position = self.ds.arr(np.array([0.6, 0.5, 0.5]), "code_length")
        vol = create_volume_source(self.ds, field=self.field)
        tf = vol.transfer_function
        tf.grey_opacity = True
        sc.add_source(vol)
        sc.save(f"test_stereospherical_{self.field[1]}.png", sigma_clip=6.0)



import os
import shutil
import tempfile

import numpy as np

from yt.data_objects.level_sets.api import Clump, add_clump_info, find_clumps
from yt.data_objects.level_sets.clump_info_items import clump_info_registry
from yt.fields.derived_field import ValidateParameter
from yt.loaders import load, load_uniform_grid
from yt.testing import assert_array_equal, assert_equal, requires_file
from yt.utilities.answer_testing.framework import data_dir_load


def test_clump_finding():
    n_c = 8
    n_p = 1
    dims = (n_c, n_c, n_c)

    density = np.ones(dims)
    high_rho = 10.0
    # add a couple disconnected density enhancements
    density[2, 2, 2] = high_rho
    density[6, 6, 6] = high_rho

    # put a particle at the center of one of them
    dx = 1.0 / n_c
    px = 2.5 * dx * np.ones(n_p)

    data = {
        "density": density,
        "particle_mass": np.ones(n_p),
        "particle_position_x": px,
        "particle_position_y": px,
        "particle_position_z": px,
    }

    ds = load_uniform_grid(data, dims)

    ad = ds.all_data()
    master_clump = Clump(ad, ("gas", "density"))
    master_clump.add_validator("min_cells", 1)

    def _total_volume(clump):
        total_vol = clump.data.quantities.total_quantity(
            [("index", "cell_volume")]
        ).in_units("cm**3")
        return "Cell Volume: %6e cm**3.", total_vol

    add_clump_info("total_volume", _total_volume)
    master_clump.add_info_item("total_volume")

    find_clumps(master_clump, 0.5, 2.0 * high_rho, 10.0)

    # there should be two children
    assert_equal(len(master_clump.children), 2)

    leaf_clumps = master_clump.leaves

    for l in leaf_clumps:
        keys = l.info.keys()
        assert "total_cells" in keys
        assert "cell_mass" in keys
        assert "max_grid_level" in keys
        assert "total_volume" in keys

    # two leaf clumps
    assert_equal(len(leaf_clumps), 2)

    # check some clump fields
    assert_equal(master_clump.children[0][("gas", "density")][0].size, 1)
    assert_equal(
        master_clump.children[0][("gas", "density")][0], ad[("gas", "density")].max()
    )
    assert_equal(master_clump.children[0][("all", "particle_mass")].size, 1)
    assert_array_equal(
        master_clump.children[0][("all", "particle_mass")], ad[("all", "particle_mass")]
    )
    assert_equal(master_clump.children[1][("gas", "density")][0].size, 1)
    assert_equal(
        master_clump.children[1][("gas", "density")][0], ad[("gas", "density")].max()
    )
    assert_equal(master_clump.children[1][("all", "particle_mass")].size, 0)

    # clean up global registry to avoid polluting other tests
    del clump_info_registry["total_volume"]


i30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(i30)
def test_clump_tree_save():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    ds = data_dir_load(i30)
    data_source = ds.disk([0.5, 0.5, 0.5], [0.0, 0.0, 1.0], (8, "kpc"), (1, "kpc"))

    field = ("gas", "density")
    step = 2.0
    c_min = 10 ** np.floor(np.log10(data_source[field]).min())
    c_max = 10 ** np.floor(np.log10(data_source[field]).max() + 1)

    master_clump = Clump(data_source, field)
    master_clump.add_info_item("center_of_mass")
    master_clump.add_validator("min_cells", 20)

    find_clumps(master_clump, c_min, c_max, step)
    leaf_clumps = master_clump.leaves

    fn = master_clump.save_as_dataset(
        fields=[
            ("gas", "density"),
            ("index", "x"),
            ("index", "y"),
            ("index", "z"),
            ("all", "particle_mass"),
        ]
    )
    ds2 = load(fn)

    # compare clumps in the tree
    t1 = [c for c in master_clump]
    t2 = [c for c in ds2.tree]
    mt1 = ds.arr([c.info["cell_mass"][1] for c in t1])
    mt2 = ds2.arr([c["clump", "cell_mass"] for c in t2])
    it1 = np.array(np.argsort(mt1).astype(int))
    it2 = np.array(np.argsort(mt2).astype(int))
    assert_array_equal(mt1[it1], mt2[it2])

    for i1, i2 in zip(it1, it2):
        ct1 = t1[i1]
        ct2 = t2[i2]
        assert_array_equal(ct1["gas", "density"], ct2["grid", "density"])
        assert_array_equal(ct1["all", "particle_mass"], ct2["all", "particle_mass"])

    # compare leaf clumps
    c1 = [c for c in leaf_clumps]
    c2 = [c for c in ds2.leaves]
    mc1 = ds.arr([c.info["cell_mass"][1] for c in c1])
    mc2 = ds2.arr([c["clump", "cell_mass"] for c in c2])
    ic1 = np.array(np.argsort(mc1).astype(int))
    ic2 = np.array(np.argsort(mc2).astype(int))
    assert_array_equal(mc1[ic1], mc2[ic2])

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


i30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(i30)
def test_clump_field_parameters():
    """
    Make sure clump finding on fields with field parameters works.
    """

    def _also_density(field, data):
        factor = data.get_field_parameter("factor")
        return factor * data[("gas", "density")]

    ds = data_dir_load(i30)
    ds.add_field(
        ("gas", "also_density"),
        function=_also_density,
        units=ds.fields.gas.density.units,
        sampling_type="cell",
        validators=[ValidateParameter("factor")],
    )
    data_source = ds.disk([0.5, 0.5, 0.5], [0.0, 0.0, 1.0], (8, "kpc"), (1, "kpc"))
    data_source.set_field_parameter("factor", 1)

    step = 2.0
    field = ("gas", "density")
    c_min = 10 ** np.floor(np.log10(data_source[field]).min())
    c_max = 10 ** np.floor(np.log10(data_source[field]).max() + 1)

    master_clump_1 = Clump(data_source, ("gas", "density"))
    master_clump_1.add_validator("min_cells", 20)
    master_clump_2 = Clump(data_source, ("gas", "also_density"))
    master_clump_2.add_validator("min_cells", 20)

    find_clumps(master_clump_1, c_min, c_max, step)
    find_clumps(master_clump_2, c_min, c_max, step)
    leaf_clumps_1 = master_clump_1.leaves
    leaf_clumps_2 = master_clump_2.leaves

    for c1, c2 in zip(leaf_clumps_1, leaf_clumps_2):
        assert_array_equal(c1["gas", "density"], c2["gas", "density"])



import inspect
from collections import Counter
from functools import reduce
from typing import List, Optional, Type


def find_lowest_subclasses(
    candidates: List[Type], *, hint: Optional[str] = None
) -> List[Type]:
    """
    This function takes a list of classes, and returns only the ones that are
    are not super classes of any others in the list. i.e. the ones that are at
    the bottom of the specified mro of classes.

    Parameters
    ----------
    candidates : Iterable
        An iterable object that is a collection of classes to find the lowest
        subclass of.

    hint : str, optional
        Only keep candidates classes that have `hint` in their name (case insensitive)

    Returns
    -------
    result : list
        A list of classes which are not super classes for any others in
        candidates.
    """

    # If there is only one input, the input candidate is always the
    # lowest class
    if len(candidates) == 1:
        return candidates
    elif len(candidates) == 0:
        return []

    mros = [inspect.getmro(c) for c in candidates]

    counters = [Counter(mro) for mro in mros]

    if len(counters) == 0:
        return []

    count = reduce(lambda x, y: x + y, counters)

    retv = [x for x in count.keys() if count[x] == 1]
    if hint is not None:
        retv = [x for x in retv if hint.lower() in x.__name__.lower()]
    return retv



import os
import shutil
import tempfile
import unittest

import numpy as np

from yt.loaders import load_particles, load_uniform_grid
from yt.testing import assert_equal, assert_raises
from yt.utilities.exceptions import (
    YTInconsistentGridFieldShape,
    YTInconsistentGridFieldShapeGridDims,
    YTInconsistentParticleFieldShape,
)


class TestEmptyLoad(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

        # create 0 byte file
        open("empty_file", "a")

        # create empty directory
        os.makedirs("empty_directory")

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)


def test_dimensionless_field_units():
    Z = np.random.uniform(size=(32, 32, 32))
    d = np.random.uniform(size=(32, 32, 32))

    data = {"density": d, "metallicity": Z}

    ds = load_uniform_grid(data, (32, 32, 32))

    dd = ds.all_data()

    assert_equal(Z.max(), float(dd["stream", "metallicity"].max()))


def test_inconsistent_field_shape():
    def load_field_field_mismatch():
        d = np.random.uniform(size=(32, 32, 32))
        t = np.random.uniform(size=(32, 64, 32))
        data = {"density": d, "temperature": t}
        load_uniform_grid(data, (32, 32, 32))

    assert_raises(YTInconsistentGridFieldShape, load_field_field_mismatch)

    def load_field_grid_mismatch():
        d = np.random.uniform(size=(32, 32, 32))
        t = np.random.uniform(size=(32, 32, 32))
        data = {"density": d, "temperature": t}
        load_uniform_grid(data, (32, 64, 32))

    assert_raises(YTInconsistentGridFieldShapeGridDims, load_field_grid_mismatch)

    def load_particle_fields_mismatch():
        x = np.random.uniform(size=100)
        y = np.random.uniform(size=100)
        z = np.random.uniform(size=200)
        data = {
            "particle_position_x": x,
            "particle_position_y": y,
            "particle_position_z": z,
        }
        load_particles(data)

    assert_raises(YTInconsistentParticleFieldShape, load_particle_fields_mismatch)


def test_parameters():
    # simple test to check that we can pass in parameters
    Z = np.random.uniform(size=(32, 32, 32))
    d = np.random.uniform(size=(32, 32, 32))

    data = {"density": d, "metallicity": Z}

    ds = load_uniform_grid(data, (32, 32, 32), parameters={"metadata_is_nice": True})
    assert ds.parameters["metadata_is_nice"]



gadget_header_specs = dict(
    default=(
        ("Npart", 6, "i"),
        ("Massarr", 6, "d"),
        ("Time", 1, "d"),
        ("Redshift", 1, "d"),
        ("FlagSfr", 1, "i"),
        ("FlagFeedback", 1, "i"),
        ("Nall", 6, "i"),
        ("FlagCooling", 1, "i"),
        ("NumFiles", 1, "i"),
        ("BoxSize", 1, "d"),
        ("Omega0", 1, "d"),
        ("OmegaLambda", 1, "d"),
        ("HubbleParam", 1, "d"),
        ("FlagAge", 1, "i"),
        ("FlagMetals", 1, "i"),
        ("NallHW", 6, "i"),
        ("unused", 16, "i"),
    ),
    pad32=(("empty", 32, "c"),),
    pad64=(("empty", 64, "c"),),
    pad128=(("empty", 128, "c"),),
    pad256=(("empty", 256, "c"),),
)

gadget_ptype_specs = dict(default=("Gas", "Halo", "Disk", "Bulge", "Stars", "Bndry"))

gadget_field_specs = dict(
    default=(
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("SmoothingLength", "Gas"),
    ),
    agora_unlv=(
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("Electron_Number_Density", "Gas"),
        ("HI_NumberDensity", "Gas"),
        ("SmoothingLength", "Gas"),
    ),
    group0000=(
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        "Potential",
        ("Temperature", "Gas"),
        ("Density", "Gas"),
        ("ElectronNumberDensity", "Gas"),
        ("HI_NumberDensity", "Gas"),
        ("SmoothingLength", "Gas"),
        ("StarFormationRate", "Gas"),
        ("DelayTime", "Gas"),
        ("FourMetalFractions", ("Gas", "Stars")),
        ("MaxTemperature", ("Gas", "Stars")),
        ("NStarsSpawned", ("Gas", "Stars")),
        ("StellarAge", "Stars"),
    ),
    magneticum_box2_hr=(
        "Coordinates",
        "Velocities",
        "ParticleIDs",
        "Mass",
        ("InternalEnergy", "Gas"),
        ("Density", "Gas"),
        ("SmoothingLength", "Gas"),
        ("ColdFraction", "Gas"),
        ("Temperature", "Gas"),
        ("StellarAge", ("Stars", "Bndry")),
        "Potential",
        ("InitialMass", "Stars"),
        ("ElevenMetalMasses", ("Gas", "Stars")),
        ("StarFormationRate", "Gas"),
        ("TrueMass", "Bndry"),
        ("AccretionRate", "Bndry"),
    ),
)

gadget_hdf5_ptypes = (
    "PartType0",
    "PartType1",
    "PartType2",
    "PartType3",
    "PartType4",
    "PartType5",
)

SNAP_FORMAT_2_OFFSET = 16



import glob
import os
import struct

import numpy as np

from yt.data_objects.static_output import ParticleFile
from yt.frontends.sph.data_structures import SPHDataset, SPHParticleIndex
from yt.utilities.cosmology import Cosmology
from yt.utilities.physical_constants import G
from yt.utilities.physical_ratios import cm_per_kpc

from .fields import TipsyFieldInfo


class TipsyFile(ParticleFile):
    def __init__(self, ds, io, filename, file_id, range=None):
        super().__init__(ds, io, filename, file_id, range)
        if not hasattr(io, "_field_list"):
            io._create_dtypes(self)
            # Check automatically what the domain size is
            io._update_domain(self)
        self._calculate_offsets(io._field_list)

    def _calculate_offsets(self, field_list, pcounts=None):
        self.field_offsets = self.io._calculate_particle_offsets(self, None)


class TipsyDataset(SPHDataset):
    _index_class = SPHParticleIndex
    _file_class = TipsyFile
    _field_info_class = TipsyFieldInfo
    _particle_mass_name = "Mass"
    _particle_coordinates_name = "Coordinates"
    _sph_ptypes = ("Gas",)
    _header_spec = (
        ("time", "d"),
        ("nbodies", "i"),
        ("ndim", "i"),
        ("nsph", "i"),
        ("ndark", "i"),
        ("nstar", "i"),
        ("dummy", "i"),
    )

    def __init__(
        self,
        filename,
        dataset_type="tipsy",
        field_dtypes=None,
        unit_base=None,
        parameter_file=None,
        cosmology_parameters=None,
        index_order=None,
        index_filename=None,
        kdtree_filename=None,
        kernel_name=None,
        bounding_box=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        # Because Tipsy outputs don't have a fixed domain boundary, one can
        # specify a bounding box which effectively gives a domain_left_edge
        # and domain_right_edge
        self.bounding_box = bounding_box
        self.filter_bbox = bounding_box is not None
        if field_dtypes is None:
            field_dtypes = {}
        success, self.endian = self._validate_header(filename)
        if not success:
            print("SOMETHING HAS GONE WRONG.  NBODIES != SUM PARTICLES.")
            print(
                "%s != (sum == %s + %s + %s)"
                % (
                    self.parameters["nbodies"],
                    self.parameters["nsph"],
                    self.parameters["ndark"],
                    self.parameters["nstar"],
                )
            )
            print("Often this can be fixed by changing the 'endian' parameter.")
            print("This defaults to '>' but may in fact be '<'.")
            raise RuntimeError
        self.storage_filename = None

        # My understanding is that dtypes are set on a field by field basis,
        # not on a (particle type, field) basis
        self._field_dtypes = field_dtypes

        self._unit_base = unit_base or {}

        self._cosmology_parameters = cosmology_parameters
        if parameter_file is not None:
            parameter_file = os.path.abspath(parameter_file)
        self._param_file = parameter_file
        filename = os.path.abspath(filename)
        if units_override is not None:
            raise RuntimeError(
                "units_override is not supported for TipsyDataset. "
                + "Use unit_base instead."
            )
        super().__init__(
            filename,
            dataset_type=dataset_type,
            unit_system=unit_system,
            index_order=index_order,
            index_filename=index_filename,
            kdtree_filename=kdtree_filename,
            kernel_name=kernel_name,
            default_species_fields=default_species_fields,
        )

    def __str__(self):
        return os.path.basename(self.parameter_filename)

    def _parse_parameter_file(self):

        # Parsing the header of the tipsy file, from this we obtain
        # the snapshot time and particle counts.

        f = open(self.parameter_filename, "rb")
        hh = self.endian + "".join("%s" % (b) for a, b in self._header_spec)
        hvals = {
            a: c
            for (a, b), c in zip(
                self._header_spec, struct.unpack(hh, f.read(struct.calcsize(hh)))
            )
        }
        self.parameters.update(hvals)
        self._header_offset = f.tell()

        # These are always true, for now.
        self.dimensionality = 3
        self.refine_by = 2
        self.parameters["HydroMethod"] = "sph"

        # Read in parameter file, if available.
        if self._param_file is None:
            pfn = glob.glob(os.path.join(self.directory, "*.param"))
            assert len(pfn) < 2, "More than one param file is in the data directory"
            if pfn == []:
                pfn = None
            else:
                pfn = pfn[0]
        else:
            pfn = self._param_file

        if pfn is not None:
            for line in (l.strip() for l in open(pfn)):
                # skip comment lines and blank lines
                l = line.strip()
                if l.startswith("#") or l == "":
                    continue
                # parse parameters according to tipsy parameter type
                param, val = (i.strip() for i in line.split("=", 1))
                val = val.split("#")[0]
                if param.startswith("n") or param.startswith("i"):
                    val = int(val)
                elif param.startswith("d"):
                    val = float(val)
                elif param.startswith("b"):
                    val = bool(float(val))
                self.parameters[param] = val

        self.current_time = hvals["time"]
        self.domain_dimensions = np.ones(3, "int32")
        periodic = self.parameters.get("bPeriodic", True)
        period = self.parameters.get("dPeriod", None)
        self._periodicity = (periodic, periodic, periodic)
        self.cosmological_simulation = float(
            self.parameters.get("bComove", self._cosmology_parameters is not None)
        )
        if self.cosmological_simulation and period is None:
            period = 1.0
        if self.bounding_box is None:
            if periodic and period is not None:
                # If we are periodic, that sets our domain width to
                # either 1 or dPeriod.
                self.domain_left_edge = np.zeros(3, "float64") - 0.5 * period
                self.domain_right_edge = np.zeros(3, "float64") + 0.5 * period
            else:
                self.domain_left_edge = None
                self.domain_right_edge = None
        else:
            # This ensures that we know a bounding box has been applied
            self._domain_override = True
            bbox = np.array(self.bounding_box, dtype="float64")
            if bbox.shape == (2, 3):
                bbox = bbox.transpose()
            self.domain_left_edge = bbox[:, 0]
            self.domain_right_edge = bbox[:, 1]

        # If the cosmology parameters dictionary got set when data is
        # loaded, we can assume it's a cosmological data set
        if self.cosmological_simulation == 1.0:
            cosm = self._cosmology_parameters or {}
            # In comoving simulations, time stores the scale factor a
            self.scale_factor = hvals["time"]
            dcosm = dict(
                current_redshift=(1.0 / self.scale_factor) - 1.0,
                omega_lambda=self.parameters.get(
                    "dLambda", cosm.get("omega_lambda", 0.0)
                ),
                omega_matter=self.parameters.get(
                    "dOmega0", cosm.get("omega_matter", 0.0)
                ),
                hubble_constant=self.parameters.get(
                    "dHubble0", cosm.get("hubble_constant", 1.0)
                ),
            )
            for param in dcosm.keys():
                pval = dcosm[param]
                setattr(self, param, pval)
        else:
            kpc_unit = self.parameters.get("dKpcUnit", 1.0)
            self._unit_base["cm"] = 1.0 / (kpc_unit * cm_per_kpc)

        self.filename_template = self.parameter_filename
        self.file_count = 1

        f.close()

    def _set_derived_attrs(self):
        if self.bounding_box is None and (
            self.domain_left_edge is None or self.domain_right_edge is None
        ):
            self.domain_left_edge = np.array([np.nan, np.nan, np.nan])
            self.domain_right_edge = np.array([np.nan, np.nan, np.nan])
            self.index
        super()._set_derived_attrs()

    def _set_code_unit_attributes(self):
        # First try to set units based on parameter file
        if self.cosmological_simulation:
            mu = self.parameters.get("dMsolUnit", 1.0)
            self.mass_unit = self.quan(mu, "Msun")
            lu = self.parameters.get("dKpcUnit", 1000.0)
            # In cosmological runs, lengths are stored as length*scale_factor
            self.length_unit = self.quan(lu, "kpc") * self.scale_factor
            density_unit = self.mass_unit / (self.length_unit / self.scale_factor) ** 3
            if "dHubble0" in self.parameters:
                # Gasoline's internal hubble constant, dHubble0, is stored in
                # units of proper code time
                self.hubble_constant *= np.sqrt(G * density_unit)
                # Finally, we scale the hubble constant by 100 km/s/Mpc
                self.hubble_constant /= self.quan(100, "km/s/Mpc")
                # If we leave it as a YTQuantity, the cosmology object
                # used below will add units back on.
                self.hubble_constant = self.hubble_constant.to_value("")
        else:
            mu = self.parameters.get("dMsolUnit", 1.0)
            self.mass_unit = self.quan(mu, "Msun")
            lu = self.parameters.get("dKpcUnit", 1.0)
            self.length_unit = self.quan(lu, "kpc")

        # If unit base is defined by the user, override all relevant units
        if self._unit_base is not None:
            for my_unit in ["length", "mass", "time"]:
                if my_unit in self._unit_base:
                    my_val = self._unit_base[my_unit]
                    my_val = (
                        self.quan(*my_val)
                        if isinstance(my_val, tuple)
                        else self.quan(my_val)
                    )
                    setattr(self, f"{my_unit}_unit", my_val)

        # Finally, set the dependent units
        if self.cosmological_simulation:
            cosmo = Cosmology(
                hubble_constant=self.hubble_constant,
                omega_matter=self.omega_matter,
                omega_lambda=self.omega_lambda,
            )
            self.current_time = cosmo.lookback_time(self.current_redshift, 1e6)
            # mass units are rho_crit(z=0) * domain volume
            mu = (
                cosmo.critical_density(0.0)
                * (1 + self.current_redshift) ** 3
                * self.length_unit**3
            )
            self.mass_unit = self.quan(mu.in_units("Msun"), "Msun")
            density_unit = self.mass_unit / (self.length_unit / self.scale_factor) ** 3
            # need to do this again because we've modified the hubble constant
            self.unit_registry.modify("h", self.hubble_constant)
        else:
            density_unit = self.mass_unit / self.length_unit**3

        if not hasattr(self, "time_unit"):
            self.time_unit = 1.0 / np.sqrt(density_unit * G)

    @staticmethod
    def _validate_header(filename):
        """
        This method automatically detects whether the tipsy file is big/little endian
        and is not corrupt/invalid.  It returns a tuple of (Valid, endianswap) where
        Valid is a boolean that is true if the file is a tipsy file, and endianswap is
        the endianness character '>' or '<'.
        """
        try:
            f = open(filename, "rb")
        except Exception:
            return False, 1
        try:
            f.seek(0, os.SEEK_END)
            fs = f.tell()
            f.seek(0, os.SEEK_SET)
            # Read in the header
            t, n, ndim, ng, nd, ns = struct.unpack("<diiiii", f.read(28))
        except (OSError, struct.error):
            return False, 1
        endianswap = "<"
        # Check Endianness
        if ndim < 1 or ndim > 3:
            endianswap = ">"
            f.seek(0)
            t, n, ndim, ng, nd, ns = struct.unpack(">diiiii", f.read(28))
        # File is borked if this is true.  The header is 28 bytes, and may
        # Be followed by a 4 byte pad.  Next comes gas particles, which use
        # 48 bytes, followed by 36 bytes per dark matter particle, and 44 bytes
        # per star particle.  If positions are stored as doubles, each of these
        # sizes is increased by 12 bytes.
        if (
            fs != 28 + 48 * ng + 36 * nd + 44 * ns
            and fs != 28 + 60 * ng + 48 * nd + 56 * ns
            and fs != 32 + 48 * ng + 36 * nd + 44 * ns
            and fs != 32 + 60 * ng + 48 * nd + 56 * ns
        ):
            f.close()
            return False, 0
        f.close()
        return True, endianswap

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return TipsyDataset._validate_header(filename)[0]
